/* images.c 
 * Copyright (C) 2001 by The matterial Developers 
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>
#include <math.h>

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "lut.h"
#include "opts.h"

#include "read_ppm.h"
#include "read_mayaiff.h"
#include "read_rla.h"
#include "read_cin.h"
#include "read_sgi.h"

#ifdef HAVE_LIBTIFF
#include "read_tiff.h"
#endif

/* this is a list of all the reader and recoginze functions
 * the recognizer should be listed first, and have the 
 * prototype:
 *
 * 	int my_rec_func(char *filename, int argc, char **argv)
 *
 * the file should exist. it should return 0 on
 * failure and 1 on sucess, that is, if the reader func 
 * specified afterward will recongnize the given file.
 *
 * the reader func should come next, and have the prototype
 *
 * 	int my_read_func(char *filename, int *w, int *h, unsigned char **buf,
 *			int argc, char **argv, lut_t *lut)
 *
 * it should:
 * 	- set w and h
 * 	- alloc buf accordingly (4*w*h bytes) [r, g, b, a] for LE
 *                                        [a, b, g, r] for BE
 * 	- read the file and put the rgb contents into buf tightly packed
 * 	- return 0 on failure, 1 on sucess
 */
 			
void * rec_and_read[] =
{
	RecMIFF, 	ReadMIFF,		/* maya iff reader */
	RecPPM,  	ReadPPM,		/* ppm reader */
	RecRLA,  	ReadRLA,		/* rla reader */
	RecCinDpx,  ReadCinDpx,		/* cineon/dpx reader */
	RecSGI, 	ReadSGI			/* SGI RGB IRIS reader */

/* the tiff stuff should always be last, as libtiff likes 
 * to puke all over stderr when it cant rec a file */
#ifdef HAVE_LIBTIFF
	,RecTIFF, 	ReadTIFF 		/* tiff reader, depends on libtiff */
#endif
};

#define PREFERRED_SCREEN_WIDTH	1920
#define PREFERRED_SCREEN_HEIGHT 1200

int nreaders = sizeof(rec_and_read) / (2*sizeof(void *));

/*==================================================*/
/* returns the number of files read sucessfully     */
int ReadFiles(char **file_list, int fileno, int *w, int *h, 
					unsigned char **image_list, int argc, char **argv, lut_t *lut)
{
	int a, b, x, y, nw, nh;
	int p00, p01, p10, p11;
	float xs, ys, dx, dy;
	unsigned char *new_buf, *old_buf;
	float sx, sy;
	int (*rec)(char * filename, int argc, char **argv);
	int (*read)(char * filename, int *w, int *h, unsigned char **buf, int argc, char ** argv,
								lut_t *lut);
	FILE *fid;

	/* parse out the scale factors here */
	sx = 1.0;
	sy = 1.0;
	optind = 1;
	while ((a = getopt(argc, argv, OPT_VALUES)) != -1)
	{
		switch(a)	
		{
			case 'x':
				sx = atof(optarg);
				break;

			case 'y':
				sy = atof(optarg);
				break;

			case 'S':
				sy = sx = atof(optarg);
				break;

			default:
				if (!index(OPT_VALUES, a))
				{
					usage(argv);
					exit(1);	
				}
				break;
		}
	}

	/* first see if the file even exists */
	fid = fopen(file_list[fileno], "r");
	if (!fid)	
	{
		fprintf(stderr, "cannot open %s\n", file_list[fileno]);
		return 0;
	}
	fclose(fid);

	for (a=0; a<nreaders; a++)
	{
		rec  = rec_and_read[2*a];
		if ((rec)(file_list[fileno], argc, argv))
		{
			read = rec_and_read[2*a+1];
			if ((read)(file_list[fileno], w, h, &image_list[fileno], argc, argv, lut))
			{
				/* overwrite scale factors, if image is too big */
				while(*w*sx > PREFERRED_SCREEN_WIDTH || *h*sy > PREFERRED_SCREEN_HEIGHT) {
					sx *= 0.5;
					sy *= 0.5;
				}
				
				/* fix the file if we have to resize */
				if (sx*sy != 1.0)
				{
					unsigned char *np;
					nw = *w;
					nh = *h;
					nw *= sx;
					nh *= sy;

					old_buf = image_list[fileno];
					np = new_buf = (unsigned char *)malloc(4*nw*nh);
					
					/* special cases for sizes: 1/2, 1/3, 1/4 and scope */
					
					if( (sx==sy) && (sx > 0.499) && (sx < 0.501) ) {
						const unsigned int step = 2*4;	
						for (y=0; y<nh; y++) {
							p00 = ((y*2)*(*w))*4;
							p10 = p00+(*w*4);
							for (x=0; x<nw; x++) {
								for (b=0; b<4; b++) {	// sum a 2x2 block
									*(np++) = (unsigned char)(((unsigned int)
													old_buf[p00+  b] + 
													old_buf[p00+4+b] + 
													old_buf[p10+  b] + 
													old_buf[p10+4+b] 
											 	) >> 2);
								}
								p00 += step;
								p10 += step;
							}
						}
					} else if( (sx==sy) && (sx < 0.334) && (sx >= 0.33) ) {
						const unsigned int step = 3*4;	
						for (y=0; y<nh; y++) {
							p01 = (p10 = (p00 = ((y*3)*(*w))*4) +(*w*4)) +(*w*4);
							for (x=0; x<nw; x++) {
								for (b=0; b<4; b++) {	// sum a 3x3 block
									*(np++) = (unsigned char)(((unsigned int)
													old_buf[p00+   b] + 
													old_buf[p00+ 4+b] + 
													old_buf[p00+ 8+b] + 
													old_buf[p01+   b] + 
													old_buf[p01+ 4+b] + 
													old_buf[p01+ 8+b] + 
													old_buf[p10+   b] + 
													old_buf[p10+ 4+b] + 
													old_buf[p10+ 8+b]
											 	) / 9);
								}
								p00 += step;
								p10 += step;
								p01 += step;
							}
						}
					} else if( (sx==sy) && (sx > 0.249) && (sx < 0.251)) {
						const unsigned int step = 4*4;	
						for (y=0; y<nh; y++) {
							p11 = (p01 = (p10 = (p00 = ((y*4)*(*w))*4) +(*w*4)) +(*w*4)) +(*w*4);
							for (x=0; x<nw; x++) {
								for (b=0; b<4; b++) {	// sum a 4x4 block
									*(np++) = (unsigned char)(((unsigned int)
													old_buf[p00+   b] + 
													old_buf[p00+ 4+b] + 
													old_buf[p00+ 8+b] + 
													old_buf[p00+12+b] + 
													old_buf[p01+   b] + 
													old_buf[p01+ 4+b] + 
													old_buf[p01+ 8+b] + 
													old_buf[p01+12+b] + 
													old_buf[p10+   b] + 
													old_buf[p10+ 4+b] + 
													old_buf[p10+ 8+b] + 
													old_buf[p10+12+b] + 
													old_buf[p11+   b] + 
													old_buf[p11+ 4+b] + 
													old_buf[p11+ 8+b] + 
													old_buf[p11+12+b] 
											 	) >> 4);
								}
								p00 += step;
								p01 += step;
								p10 += step;
								p11 += step;
							}
						}
					} else if( (sx > 0.499) && (sx < 0.501) && (sy > 0.249) && (sy < 0.251)) {
						/* x=0.5, y=0.25 -> unsqueeze scope */
						const unsigned int step = 2*4;	
						for (y=0; y<nh; y++) {
							p00 = ((y*4)*(*w))*4;
							p01 = p00+(*w*4);
							p10 = p01+(*w*4);
							p11 = p10+(*w*4);
							for (x=0; x<nw; x++) {
								for (b=0; b<4; b++) {	// sum a 2x4 block
									*(np++) = (unsigned char)(((unsigned int)
													old_buf[p00+  b] + 
													old_buf[p00+4+b] + 
													old_buf[p01+  b] + 
													old_buf[p01+4+b] + 
													old_buf[p10+  b] + 
													old_buf[p10+4+b] + 
													old_buf[p11+  b] + 
													old_buf[p11+4+b]  
											 	) >> 3);
								}
								p00 += step;
								p01 += step;
								p10 += step;
								p11 += step;
							}
						}
					} else if( (sx == 1.0) && (sy > 0.499) && (sy < 0.501)) {
						/* x=1, y=0.5 -> unsqueeze scope */
						const unsigned int step = 4;	
						for (y=0; y<nh; y++) {
							p00 = ((y*2)*(*w))*4;
							p01 = p00+(*w*4);
							for (x=0; x<nw; x++) {
								for (b=0; b<4; b++) {	// sum a 2x1 block
									*(np++) = (unsigned char)(((unsigned int)
													old_buf[p00+b] + 
													old_buf[p01+b]
											 	) >> 1);
								}
								p00 += step;
								p01 += step;
							}
						}
					} else { /* fall back to slow scaling */
											
						/* this isnt the prettiest scaling, but i doubt
						 * anyone will be scaling great distances.. and its
						 * quick and dirty.. */
						 
						for (y=0; y<nh; y++) {
							for (x=0; x<nw; x++) {
								xs = (float)x / sx;
								ys = (float)y / sy;
								if (xs < 0)   xs = 0;
								if (xs >= *w) xs = *w-1;
								if (ys < 0)   ys = 0;
								if (ys >= *h) ys = *h-1;

								dx = xs - floor(xs);
								dy = ys - floor(ys);

								p00 = (floor(ys)*(*w)+floor(xs))*4;
								p10 = (floor(ys)*(*w)+ceil(xs))*4;
								p01 = (ceil(ys)*(*w)+floor(xs))*4;
								p11 = (ceil(ys)*(*w)+ceil(xs))*4;

								for (b=0; b<4; b++) {
									*(np++) =
										(1.-dx)*(1.-dy)*old_buf[p00+b] +
										(dx)*(1.-dy)*old_buf[p10+b] +
										(1.-dx)*(dy)*old_buf[p01+b] +
										(dx)*(dy)*old_buf[p11+b];
								}
							}
						}
					}
					free(image_list[fileno]);
					image_list[fileno] = new_buf;
					*w = nw;
					*h = nh;
				}
				return 1;
			}
			*w *= sx;
			*h *= sy;
			return 0;
		}
	}

	fprintf(stderr, "no reader found for %s\n", file_list[fileno]);
	return 0;
}
