/* read_cin.c	- read Kodak Cineon scan files version 4.5
 * Copyright (C) 2001 by The matterial Developers 
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

#include <glib.h>

#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>
#include <math.h>

#include "lut.h"
#include "opts.h"

/*
  For details on the Cineon file format look at
      http://www.cinesite.com/CineonTech/imageFileFormat4.5/imageFileFormat4.5.html
  and
      http://www.cineon.com

  For details on SMPTE DPX
  http://www.cinesite.com/CineonTech/fsImgFrmt/fsImgFrmt.html
*/

struct CIN_FILE_INFO
{   guint32 magicNum;
    guint32 imgOffset;
    guint32 genSecLen;
    guint32 mpSecLen;
    guint32 varSecLen;
    guint32 fileSize;
    gchar	hdrVersNum[8];
    gchar	imgFilename[100];
    gchar	creatDate[12];
    gchar	creatTime[12];
    gchar	reserved[36];
};

struct CIN_CHANNEL_SPECIFIER
{   guchar	channelType;
    guchar	channelMetric;
    guchar	bitsPerPixel;
    gchar	unused;
    guint32 pixelsPerLine;
    guint32 linesPerImg;
    float	minValue;
    float	minQuantity;
    float	maxValue;
    float	maxQuantity;
};

#define CHANNELS   8

struct CIN_IMAGE_INFO
{   guchar	orientation;
    guchar	numChannels;
    guint16 unused;
    struct	CIN_CHANNEL_SPECIFIER   channels[CHANNELS];
    float	whitePointX;
    float	whitePointY;
    float	primRedX;
    float	primRedY;
    float	primGrnX;
    float	primGrnY;
    float	primBluX;
    float	primBluY;
    gchar	label[200];
    gchar	reserved[28];
};

struct CIN_IMAGE_DATA_FORMAT
{   guchar	interleave;
    guchar	packing;
    guchar	dataSign;
    guchar	imgSense;
    guint32 eolPadding;
    guint32 eocPadding;
    gchar	reserved[20];
};

struct CIN_IMAGE_ORIGIN
{   gint32  xOffset;
    gint32  yOffset;
    gchar	imgFilename[100];
    gchar	creatDate[12];
    gchar	creatTime[12];
    gchar	inputDevice[64];
    gchar	inDevMdlNum[32];
    gchar	inDevSrlNum[32];
    float	inDevXPitch;
    float	inDevYPitch;
    float	imgGamma;
    gchar	reserved[40];
};

struct CIN_FRAME_INFO
{   gchar	filmMfgCode;
    gchar	filmTypeCode;
    guint8  perfOffset;
    gchar	unused;
    guint32 prefix;
    guint32 count;
    gchar	format[32];
    guint32 framePosInSeq;
    float	frameRate;
    gchar	frameAttribute[32];
    gchar	slateInfo[200];
    gchar	reserved[740];
};

typedef struct st_cin_header
{   struct CIN_FILE_INFO         fileInfo;
    struct CIN_IMAGE_INFO        imgInfo;
    struct CIN_IMAGE_DATA_FORMAT imgDataFrmt;
    struct CIN_IMAGE_ORIGIN	     imgOrigin;
    struct CIN_FRAME_INFO        frameInfo;
} cin_header;


/*===============================================================*/
/* internal funcs                                                */

/* this is an internal LUT to convert from logarithmic 10bit to linear 16bit */
static gint cinDpxNeedsInit = 1;
static guint16 *log2Lin;
static gint rOffset = 0;
static gint gOffset = 0;
static gint bOffset = 0;

void initCinDpx(int argc, char **argv)
{
	unsigned int blackPoint=95, whitePoint=685, i;
	char 		*env;
	int 		o;
	int 		 pixbuf;
	float		 pixfloat, blackfix, a, b, c, gCorr=1.0;

	log2Lin = (guint16 *) malloc(2048);	
/*
	if(env=getenv(ENVPRE"MIN_LOG"))  
		blackPoint = MIN((unsigned int) atoi(env),1023);
		
	if(env=getenv(ENVPRE"MAX_LOG"))	
		whitePoint = MIN(1023,MAX((unsigned int) atoi(env),blackPoint));

	if(blackPoint!=95 || whitePoint!=685)
		fprintf(stderr, "Using non-standard log->lin range: %d-%d\n", 
			blackPoint, whitePoint);
*/	
	optind = 1;
	while ((o = getopt(argc, argv, OPT_VALUES)) != -1)
	{
		switch(o)	
		{
			case 'R':	rOffset = atoi(optarg); break;
			case 'G':	gOffset = atoi(optarg); break;
			case 'B':	bOffset = atoi(optarg); break;
			case 'D': {	
				int add = atoi(optarg);
				rOffset += add;
				gOffset += add;
				bOffset += add;
				break;
			}
		}
	}
	
	pixbuf = blackPoint - whitePoint;
	pixfloat = (float)pixbuf * 0.002048 / 0.6;
	a = powf(10.0, pixfloat);	
	blackfix = powf(a, gCorr);

	for (i = 0; i <= 1023; ++i) {
 		if (i > whitePoint) 
			pixbuf = 0;
		else if (i < blackPoint) 
			pixbuf = blackPoint - whitePoint;
		else 
			pixbuf = i - whitePoint;
		a = ((float)pixbuf) * 0.002048 / 0.6;
		b = powf(10.0, a);
		c = powf(b, gCorr);
		pixfloat = (c - blackfix)/(1.0 - blackfix);		
		log2Lin[i] = (unsigned short) (pixfloat*65535.0);
	}

/* todo: maybe we want a soft knee for the highlights
	if( smoothStart < 1023 ) {
		// now smooth it according to Ians approach
		for (i = 0; i <= 1023; ++i) {
			if ( i < smoothStart ) // not into smooth range
				smoothLut[i] = lutFloat[i];
			else {
				if (range < ((1023 - whitePoint)/2))  { // ramp in
					rangeFloat = rangeFloat + 0.5;
					range = (int)rangeFloat;
				}
				count = 0; value = 0;
				for(lut_ptr = i-range; lut_ptr <= i+range; ++lut_ptr) {
					if (lut_ptr >= 0 && lut_ptr <= 1023) {
						count++;
						value = value + lutFloat[lut_ptr];
					} 
				}
				smoothLut[i] = (value/(float)count);
			}
		}

		// little rounding error weirdness ...
		for (i = 0; i <= 1023; ++i) {
			lut10[i] = (guint16)( smoothLut[i] * 65535.01 );
		}
	}
*/
}

/* 0 on sucess, 1 on failure, read and fix order in a header     */
int readCinDpxHeader(
	FILE *fid, 
	int *w, int *h, 
	guint32 *imgOffset, guint32 *eolPadding, 
	int *doSwap )
{
	guint32 magic = 0;
	int topDown;
	cin_header cin;
	guint16	orientation, elements, packing, encoding;
	guint8	bits;

	*doSwap = 0;
		
	fread(&magic, 1, sizeof(magic), fid);
	
	switch(magic) {

		/* CIN */
		case 0xd75f2a80L : 
			*doSwap = 1;
		case 0x802a5fd7L : 
		
			if (fread(&cin.fileInfo.imgOffset, 1, sizeof(cin_header)-sizeof(magic), fid) != sizeof(cin_header)-sizeof(magic))
				return 1;
	
			if(*doSwap) {
				/* we could byte-swap ALL header fields,
				 * but honestly most of them are pretty uninteresting
				 * so we only treat the important ones.
				 * If you feel like interpreting more, they're all defined above */
	
				cin.fileInfo.magicNum = magic;
				cin.fileInfo.imgOffset = GUINT32_SWAP_LE_BE(cin.fileInfo.imgOffset);
				cin.imgInfo.channels[0].pixelsPerLine = GUINT32_SWAP_LE_BE(cin.imgInfo.channels[0].pixelsPerLine);
				cin.imgInfo.channels[0].linesPerImg = GUINT32_SWAP_LE_BE(cin.imgInfo.channels[0].linesPerImg);
				cin.imgDataFrmt.eolPadding = GUINT32_SWAP_LE_BE(cin.imgDataFrmt.eolPadding);
				cin.imgDataFrmt.eocPadding = GUINT32_SWAP_LE_BE(cin.imgDataFrmt.eocPadding);
			}
			
			if(strcmp(cin.fileInfo.hdrVersNum, "V4.5\0"))
				return 1;
			
			switch(cin.imgInfo.orientation) {
				case 0: topDown = 1; break;	// left to right - top to bottom
				case 1: topDown = 0; break; // left to right - bottom to top
				default :
					fprintf(stderr, "WARNING - unsupported CIN image orientation: %d, assuming l2r t2b\n",
						cin.imgInfo.orientation );
					topDown = 0;	// keep going
			}

			if(cin.imgInfo.numChannels != 3) {
				fprintf(stderr, "Unsupported number of channels in CIN: %d (should be 3=RGB)\n",
					cin.imgInfo.numChannels);
				return 1;
			}
	
			if(cin.imgInfo.channels[0].bitsPerPixel != 10) {
				fprintf(stderr, "Unsupported bits per sample in CIN: %d (should be 10)\n",
					cin.imgInfo.channels[0].bitsPerPixel );
				return 1;
			}
	
			if(cin.imgDataFrmt.interleave != 0) {
				fprintf(stderr, "Unsupported data interleave in CIN: %d (should be 0=pixel interleave)\n",
					cin.imgDataFrmt.interleave);
				return 1;
			}
		
			*w = cin.imgInfo.channels[0].pixelsPerLine;
			*h = cin.imgInfo.channels[0].linesPerImg;
			*eolPadding = cin.imgDataFrmt.eolPadding;
			*imgOffset = cin.fileInfo.imgOffset;
			
			break;
			
		/* DPX */
		case 0x58504453L : 
			*doSwap = 1;
		case 0x53445058L : 
		
			fseek(fid, 768, SEEK_SET);
			fread(&orientation, 1, 2, fid); 
			if(*doSwap) orientation = GUINT16_SWAP_LE_BE(orientation);
			switch(orientation) {
				case 0: topDown = 1; break;	// left to right - top to bottom
				case 2: topDown = 0; break; // left to right - bottom to top
				default :
					fprintf(stderr,
						"WARNING - unsupported DPX image orientation: %d\n",
						orientation);
					topDown = 0;	// keep going ;-)
			}
			
			fread(&elements, 1, 2, fid);	// number of elements (1-8; 1=RGB combined)
			if(*doSwap) elements = GUINT16_SWAP_LE_BE(elements);
			if(elements != 1) {
				fprintf(stderr,
					"Unsupported number of image elements in DPX: %d (should be 1=combined RGB)\n",
					elements);
				return 0;
			}

			fread(w, 1, 4, fid);	// width
			if(*doSwap) *w = GUINT32_SWAP_LE_BE(*w);
			fread(h, 1, 4, fid);	// height
			if(*doSwap) *h = GUINT32_SWAP_LE_BE(*h);

			fseek(fid, 803, SEEK_SET);	// Bit size
			bits = fgetc(fid);
			if(bits != 10) {
				fprintf(stderr,
					"Unsupported bits per sample in DPX: %d (should be 10)\n",
					bits);
				return 0;
			}

			fread(&packing, 1, 2, fid);	// Packing	(0=packed into 32bit words, 1= filled to 32bit)
			if(*doSwap) packing = GUINT16_SWAP_LE_BE(packing);
			if(packing != 1) {
				fprintf(stderr,
					"Unsupported 32bit packing %d (must say 1=filled)\n",
					packing);
				return 0;
			}
	
			fread(&encoding, 1, 2, fid);	// Encoding (0=no encoding, 1=RLE)
			if(*doSwap) encoding = GUINT16_SWAP_LE_BE(encoding);
			if(encoding != 0) {
				fprintf(stderr,
					"Unsupported encoding %d (must say 0=no encoding)\n",
					encoding);
				return 0;
			}
			
			fread(imgOffset, 1, 4, fid);
			if(*doSwap) *imgOffset = GUINT32_SWAP_LE_BE(*imgOffset);
			
			*eolPadding = 0;	/*!! to do */
			
			break;
			
		default : 
			return 1;	// still invalid BE or LE magic
	}
		
	return 0;
}

/*===============================================================*/
/* functions available to other                                  */

void UsageCinDpx()
{
	printf("\n\tNegative density corrections for CIN/DPX scans (best use w/ -g 2.5) :\n");
	printf("\t  -R n\t\tcorrect red by +/- n (90 = 1 fstop)\n");
	printf("\t  -G n\t\tsame for green\n");
	printf("\t  -B n\t\tsame for blue\n");
	printf("\t  -D n\t\tapply to red/green/blue at once\n");
}

/* we are passed the filename of a file which is known to be
 * readable. we are to look at it to see if we recognize the file.
 *
 * return 0 if we cannot recognize the file */
 
int RecCinDpx(char *filename, int argc, char **argv)
{
	FILE *fid;
	int doSwap = 0, w, h, imgOffset, eolPadding;

	fid = fopen(filename, "r");
	if (readCinDpxHeader(fid, &w, &h, &imgOffset, &eolPadding, &doSwap))
	{
		fclose(fid);
		return 0;
	}
	fclose(fid);
	
	return 1;	/* 1 -- the file provides rgb */
}

/* finally load the image */
int ReadCinDpx(
		char *filename, 
		int *w, int *h, 
		unsigned char **buf, 
		int argc, char **argv,
		lut_t *lut
	) 
{
	FILE *fid;
	int doSwap = 0, line, pix, lineLen;
	guint32 log, eolPadding, imgOffset, *lineBuf;
	guchar *p;
	
	if(cinDpxNeedsInit) {
		initCinDpx(argc, argv);
		cinDpxNeedsInit = 0;
	}

	fid = fopen(filename, "r");
	if (!fid) {
		fprintf(stderr, "cannot read file %s\n", filename);
		return 0;
	}

	readCinDpxHeader(fid, w, h, &imgOffset, &eolPadding, &doSwap);

	fseek(fid, imgOffset, SEEK_SET);
	
	/* image buffer */
	p = *buf = (guchar *)malloc(4*(*w)*(*h));
	// memset(p, 0, 4*(*w)*(*h));

	/* line buffer for faster reading */
	lineLen = (sizeof(guint32)*(*w))+ eolPadding;
	lineBuf = (guint32*) malloc(lineLen);
	
	/* We do the if outside, because of performance */
	if(rOffset || gOffset || bOffset) { 
		for (line=0; line<*h; line++) {
			fread(lineBuf, 1, lineLen, fid);
			for (pix=0; pix<*w; pix++) {
				log = doSwap ? GUINT32_SWAP_LE_BE(lineBuf[pix]) : lineBuf[pix];
				if (G_BYTE_ORDER == G_LITTLE_ENDIAN) {
					*(p++) = (guchar) (lut->lut16[log2Lin[MIN(MAX(((int)(log >>  2) & 0x3ff)+rOffset, 0), 1023)]] >> 8);
					*(p++) = (guchar) (lut->lut16[log2Lin[MIN(MAX(((int)(log >> 12) & 0x3ff)+gOffset, 0), 1023)]] >> 8);
					*(p++) = (guchar) (lut->lut16[log2Lin[MIN(MAX(((int)(log >> 22) & 0x3ff)+bOffset, 0), 1023)]] >> 8);
					*(p++) = 255; 
				} else {
					*(p++) = 255; 
					*(p++) = (guchar) (lut->lut16[log2Lin[MIN(MAX(((int)(log >>  2) & 0x3ff)+rOffset, 0), 1023)]] >> 8);
					*(p++) = (guchar) (lut->lut16[log2Lin[MIN(MAX(((int)(log >> 12) & 0x3ff)+gOffset, 0), 1023)]] >> 8);
					*(p++) = (guchar) (lut->lut16[log2Lin[MIN(MAX(((int)(log >> 22) & 0x3ff)+bOffset, 0), 1023)]] >> 8);
				}
			}
		}
	} else {
		for (line=0; line<*h; line++) {
			fread(lineBuf, 1, lineLen, fid);
			for (pix=0; pix<*w; pix++) {
				log = doSwap ? GUINT32_SWAP_LE_BE(lineBuf[pix]) : lineBuf[pix];
				if (G_BYTE_ORDER == G_LITTLE_ENDIAN) {
					*(p++) = (guchar) (lut->lut16[log2Lin[(log >>  2) & 0x3ff]] >> 8); 
					*(p++) = (guchar) (lut->lut16[log2Lin[(log >> 12) & 0x3ff]] >> 8); 
					*(p++) = (guchar) (lut->lut16[log2Lin[(log >> 22) & 0x3ff]] >> 8); 
					*(p++) = 255; 
				} else {
					*(p++) = 255; 
					*(p++) = (guchar) (lut->lut16[log2Lin[(log >>  2) & 0x3ff]] >> 8); 
					*(p++) = (guchar) (lut->lut16[log2Lin[(log >> 12) & 0x3ff]] >> 8); 
					*(p++) = (guchar) (lut->lut16[log2Lin[(log >> 22) & 0x3ff]] >> 8); 
				}
			}
		}
	}
	free(lineBuf);
	fclose(fid);
	return 1;
}
