/* read_sgi.c 
 * Copyright (C) 2001 by The matterial Developers 
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

#include <stdio.h>
#include <stdlib.h>
#include <glib.h>

#include "lut.h"

/**************************************************************************
 * For details on the SGI/RGB file format see
 * http://astronomy.swin.edu.au/pbourke/dataformats/sgirgb/sgiversion.html
 *
 * Be aware, the channels are not interleaved and the order does match
 * neither SGI, nor PC frame buffers - a big mess
 * Written and tested with 8/16bit RGB[A} images by Thomas Kumlehn
 * Could not create test images for single channel or gray+alpha,
 * this may go belly up.
 **************************************************************************/

typedef struct stSgiHeader {
	guint16 	magic;		/* IRIS image file magic number = 474 */
	guchar		rle; 		/* Storage format, 0=verbatim, 1=RLE */
	guchar		bpc;		/* Number of bytes per pixel channel */
	guint16 	dimension;	/* Number of dimensions */
	guint16 	width;		/* X size in pixels */
	guint16 	height; 	/* Y size in pixels */
	guint16 	channels;	/* Number of channels */
	guint32 	pixMin; 	/* Minimum pixel value */
	guint32 	pixMax; 	/* Maximum pixel value */
	guint32 	filler1; 	/* Ignored */
	gchar		imageName;	/* original image file name */
	guint32 	colorMap;	/* Colormap ID */	
	guint32 	filler2; 	/* Ignored */
	/* header is effectively 512 bytes long, so we later use fseek */
} sgiHeader;

void expandRow8(
	guchar *optr, 
	guchar *iptr
	)
{
    guchar pixel, count;

    while(1) {
        pixel = *(iptr++);
        if ( !(count = (pixel & 0x7f)) )
            return;
        if(pixel & 0x80) {
            while(count--) 
                *(optr++) = *(iptr++);
        } else {
            pixel = *(iptr++);
            while(count--)
                *(optr++) = pixel;
        }
    }
}

void expandRow16(
	guint16 *optr,
	guint16 *iptr
	)
{
    guint16 pixel;
	guchar count;

    while(1) {
        pixel = *(iptr++);
        if ( !(count = (pixel & 0x7f)) )
            return;
        if(pixel & 0x80) {
            while(count--) 
                *(optr++) = *(iptr++);
        } else {
            pixel = *(iptr++);
            while(count--) 
                *(optr++) = pixel;
        }
    }
}

/*====================================================*/
void UsageSGI()
{
}

/*====================================================*/
int RecSGI(char *filename, int argc, char **argv)
{
	FILE *fid;
	guint16 magic;
	guchar rle;
	guchar bpc;

	fid = fopen(filename, "r");
	if (!fid) 	
	{
		fprintf(stderr, "cannot open %s\n", filename);
		return 0;
	}

	fread(&magic, 2,  1, fid);
	rle = fgetc(fid);
	bpc = fgetc(fid);
	
	fclose(fid);

	if ((rle < 2) && (bpc < 3))
		if (G_BYTE_ORDER == G_LITTLE_ENDIAN)
			return GUINT16_SWAP_LE_BE(magic) == 474;
		else
			return magic == 474;
	else
		return 0;
}

/*====================================================*/
/* returns 1 on sucess                                */
int ReadSGI(char *filename, int *w, int *h, unsigned char **buf,
		int argc, char **argv, lut_t *lut)
{
	FILE *fid;
	int size, ch, fileCh;
	guchar *cBuf, *p, *rleBuf;
	guint16 *sBuf, *sRleBuf, line, pix;
	size_t lineSize;
	long *startTab, *lenTab;
	sgiHeader sh;

	fid = fopen(filename, "r");
	if (!fid) 	
	{
		fprintf(stderr, "cannot open %s\n", filename);
		return 0;
	}

	fread(&sh, sizeof(sh), 1, fid);

	if (G_BYTE_ORDER == G_LITTLE_ENDIAN) 
	{	/* handle byte swapping for non-char fields */
		sh.magic		= GUINT16_SWAP_LE_BE(sh.magic);
		sh.dimension	= GUINT16_SWAP_LE_BE(sh.dimension);
		sh.width		= GUINT16_SWAP_LE_BE(sh.width);
		sh.height		= GUINT16_SWAP_LE_BE(sh.height);
		sh.channels		= GUINT16_SWAP_LE_BE(sh.channels);
		sh.pixMin		= GUINT32_SWAP_LE_BE(sh.pixMin);
		sh.pixMax		= GUINT32_SWAP_LE_BE(sh.pixMax);
		sh.colorMap		= GUINT32_SWAP_LE_BE(sh.colorMap);
	}
	
	if(sh.dimension != 3) {
		fclose(fid);
		return 0;
	}	
	
	*w = sh.width;
	*h = sh.height;
	size = (*w) * (*h);
	 
	*buf = (unsigned char *)malloc(4*size);

	fseek(fid, 512, SEEK_SET);	/* data or tables start here */
	
	lineSize = sh.bpc * (*w);
	
	if(sh.rle) {
		int tabSize = sh.channels * (*h);
		rleBuf = (void *) malloc(lineSize+500);	/* give it some headroom */
		sRleBuf = (guint16 *) rleBuf;
		
		startTab = (long *) malloc(sizeof(long)*tabSize);
		fread(startTab, sizeof(long)*tabSize, 1, fid);
		if (G_BYTE_ORDER == G_LITTLE_ENDIAN) {
			for(line=0; line<tabSize; line++) {
				startTab[line] = GUINT32_SWAP_LE_BE(startTab[line]);
			}
		}
		
		lenTab = (long *) malloc(sizeof(long)*tabSize);
		fread(lenTab, sizeof(long)*tabSize, 1, fid);
		if (G_BYTE_ORDER == G_LITTLE_ENDIAN) {
			for(line=0; line<tabSize; line++) {
				lenTab[line] = GUINT32_SWAP_LE_BE(lenTab[line]);
			}
		}
	}

#define CHAN_INIT(C)  if (G_BYTE_ORDER != G_LITTLE_ENDIAN)\
						switch(C) {\
							case 0: p = *buf+3; break;\
							case 1: p = *buf+2; break;\
							case 2: p = *buf+1; break;\
							case 3: p = *buf+0; break;\
						} \
					 else \
						switch(C) {\
							case 0: p = *buf+2; break;\
							case 1: p = *buf+1; break;\
							case 2: p = *buf+0; break;\
							case 3: p = *buf+3; break;\
						}

	switch (sh.bpc) {
		case 1 : /* 8 bit */
			cBuf = (gchar *) malloc(lineSize);

			if(sh.channels >= 3) {	/* RGB[A] */
				for (ch=0; ch<=3 && ch<sh.channels; ch++) {
					CHAN_INIT(ch)
					for (line=0; line<*h; line++) {
						if(sh.rle) {
							fseek(fid, startTab[(*h*(ch+1))-line-1], SEEK_SET);
							fread(rleBuf, lenTab[(*h*(ch+1))-line-1], 1, fid);
							expandRow8(cBuf, rleBuf);
						} else {
							fseek(fid, (((*h*(ch+1))-line-1)*lineSize)+512, SEEK_SET);
							fread(cBuf, lineSize, 1, fid);
						}
						for (pix=0; pix<*w; pix++) {
							*p = lut->lut8[cBuf[pix]];
							p += 4;
						}
					}
				}
				if(sh.channels == 3) {
					/* fill missing alpha */
					CHAN_INIT(3)
					for (line=0; line<*h; line++) {
						for (pix=0; pix<*w; pix++) {
							*p = 255;
							p += 4;
						}
					}
				}
			} else if(sh.channels >= 1) {	/* b/w and maybe alpha */
				CHAN_INIT(0)
				for (line=0; line<*h; line++) {
					if(sh.rle) {
						fseek(fid, startTab[*h-line-1], SEEK_SET);
						fread(rleBuf, lenTab[*h-line-1], 1, fid);
						expandRow8(cBuf, rleBuf);
					} else {
						fseek(fid, ((*h-line-1)*lineSize)+512, SEEK_SET);
						fread(cBuf, lineSize, 1, fid);
					}
					for (pix=0; pix<*w; pix++) {
						guchar val = (guchar) (lut->lut8[cBuf[pix]]);
						if (G_BYTE_ORDER == G_LITTLE_ENDIAN) {
							*(p++) = val;
							*(p++) = val;
							*(p++) = val;
							p++;
						} else {
							p++;
							*(p++) = val;
							*(p++) = val;
							*(p++) = val;
						}
					}
				}
				/* now either read or init alpha */
				CHAN_INIT(3)
				if(sh.channels == 2) {
					for (line=0; line<*h; line++) {
						if(sh.rle) {
							fseek(fid, startTab[2*(*h)-line-1], SEEK_SET);
							fread(rleBuf, lenTab[2*(*h)-line-1], 1, fid);
							expandRow8(cBuf, rleBuf);
						} else {
							fseek(fid, (((2*(*h))-line-1)*lineSize)+512, SEEK_SET);
							fread(cBuf, lineSize, 1, fid);
						}
						for (pix=0; pix<*w; pix++) {
							*p = (guchar) (lut->lut8[cBuf[pix]]);
							p += 4;
						}
					}
				} else {
					for (line=0; line<*h; line++) {
						for (pix=0; pix<*w; pix++) {
							*p = 255;
							p += 4;
						}
					}
				}
			}
			free(cBuf);
			break;

		case 2 : /* 16 bit */
			sBuf = (guint16 *) malloc(lineSize);

			if(sh.channels >= 3) {	/* RGB[A] */
				for (ch=0; ch<=3 && ch<sh.channels; ch++) {
					CHAN_INIT(ch)
					for (line=0; line<*h; line++) {
						if(sh.rle) {
							fseek(fid, startTab[(*h*(ch+1))-line-1], SEEK_SET);
							fread(sRleBuf, lenTab[(*h*(ch+1))-line-1], 1, fid);
							if (G_BYTE_ORDER == G_LITTLE_ENDIAN)
								for(pix=0; pix<lenTab[line]>>1; pix++)
									sRleBuf[pix] = GUINT16_SWAP_LE_BE(sRleBuf[pix]);
							expandRow16(sBuf, sRleBuf);
						} else {
							fseek(fid, (((*h*(ch+1))-line-1)*lineSize)+512, SEEK_SET);
							fread(sBuf, lineSize, 1, fid);
							if (G_BYTE_ORDER == G_LITTLE_ENDIAN)
								for(pix=0; pix<*w; pix++) 
									sBuf[pix] = GUINT16_SWAP_LE_BE(sBuf[pix]);
						}
						for (pix=0; pix<*w; pix++) {
							*p = (guchar)(lut->lut16[sBuf[pix]]>>8);
							p += 4;
						}
					}
				}
				if(sh.channels == 3) {
					/* fill missing alpha */
					CHAN_INIT(3)
					for (line=0; line<*h; line++) {
						for (pix=0; pix<*w; pix++) {
							*p = 255;
							p += 4;
						}
					}
				}
			} else if(sh.channels >= 1) {	/* b/w and maybe alpha */
				CHAN_INIT(0)
				for (line=0; line<*h; line++) {
					if(sh.rle) {
						fseek(fid, startTab[(*h)-line-1], SEEK_SET);
						fread(sRleBuf, lenTab[(*h)-line-1], 1, fid);
						if (G_BYTE_ORDER == G_LITTLE_ENDIAN)
							for(pix=0; pix<lenTab[line]>>1; pix++)
								sRleBuf[pix] = GUINT16_SWAP_LE_BE(sRleBuf[pix]);
						expandRow16(sBuf, sRleBuf);
					} else {
						fseek(fid, (((*h)-line-1)*lineSize)+512, SEEK_SET);
						fread(sBuf, lineSize, 1, fid);
						if (G_BYTE_ORDER == G_LITTLE_ENDIAN)
							for(pix=0; pix<*w; pix++)
								sBuf[pix] = GUINT16_SWAP_LE_BE(sBuf[pix]);
					}
					for (pix=0; pix<*w; pix++) {
						guchar cval = (guchar) (lut->lut16[sBuf[pix]]>>8);
						if (G_BYTE_ORDER == G_LITTLE_ENDIAN) {
							*(p++) = cval;
							*(p++) = cval;
							*(p++) = cval;
							p++;
						} else {
							p++;
							*(p++) = cval;
							*(p++) = cval;
							*(p++) = cval;
						}
					}
				}
				/* now either read or init alpha */
				CHAN_INIT(3);
				if(sh.channels == 2) {
					for (line=0; line<*h; line++) {
						if(sh.rle) {
							fseek(fid, startTab[2*(*h)-line-1], SEEK_SET);
							fread(sRleBuf, lenTab[2*(*h)-line-1], 1, fid);
							if (G_BYTE_ORDER == G_LITTLE_ENDIAN)
								for(pix=0; pix<lenTab[line]>>1; pix++)
									sRleBuf[pix] = GUINT16_SWAP_LE_BE(sRleBuf[pix]);
							expandRow16(sBuf, sRleBuf);
						} else {
							fseek(fid, (((2*(*h))-line-1)*lineSize)+512, SEEK_SET);
							fread(sBuf, lineSize, 1, fid);
							if (G_BYTE_ORDER == G_LITTLE_ENDIAN)
								for(pix=0; pix<*w; pix++)
									sBuf[pix] = GUINT16_SWAP_LE_BE(sBuf[pix]);
						}
						for (pix=0; pix<*w; pix++) {
							*p = (guchar) (lut->lut16[sBuf[pix]]>>8);
							p += 4;
						}
					}
				} else {
					for (line=0; line<*h; line++) {
						for (pix=0; pix<*w; pix++) {
							*p = 255;
							p += 4;
						}
					}
				}
			}
			free(sBuf);
			break;
	}
	
	if(sh.rle) {
		free(startTab);
		free(lenTab);
		free(rleBuf);
	}

	fclose(fid);

	return 1;
}
