/* timer.c 
 * Copyright (C) 2001 by The matterial Developers 
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

/* linux/intel timer dependant stuff... */

#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef X86_LINUX_TIMER
	#define my_timer unsigned long long int 
#else
	#define my_timer struct timeval
#endif

double secs_per_clock;

/*=========================================================*/
#ifdef X86_LINUX_TIMER
my_timer SecsToClocks(double secs)
{	
	return secs / secs_per_clock;
}

/*=========================================================*/
/* nasty asm to return the instruction count since reboot  */
__inline__ unsigned long long int rdtsc()
{
        unsigned long long int x;
        __asm__ volatile(".byte 0x0f, 0x31" : "=A" (x));
        return x;
}
#endif

/*=========================================================*/
/* b comes later than a, return the difference in seconds  */
double TimeDiff(struct timeval *a, struct timeval *b)
{
	int usec = b->tv_usec - a->tv_usec;

	return (b->tv_sec - a->tv_sec) + (usec / 1e6);
}

/*=========================================================*/
/* negociate the cpu speed                                 */
void TimerInit()
{
	FILE *fid;
	char word[100];
	double clocks_per_sec;

#ifdef X86_LINUX_TIMER
	fid = fopen("/proc/cpuinfo", "r");
	if (!fid)
	{
		printf("cannot open /proc/cpuinfo\n");
		exit(1);
	}

	strcpy(word, "");
	while (strcmp(word, "MHz"))
		fscanf(fid, "%s", word);
	fscanf(fid, "%s", word);
	fscanf(fid, "%s", word);

	fclose(fid);

	clocks_per_sec = atof(word) * 1000000.;
	secs_per_clock = 1. / clocks_per_sec;

//	printf("%f clocks/sec\n", clocks_per_sec);
#endif

}

/*=========================================================*/
/* start a timer.. pass a ulli to hold the start count     */
void TimerStart(my_timer *s)
{
#ifdef X86_LINUX_TIMER
	*s = rdtsc();
#else
	gettimeofday(s, NULL);
#endif
}

/*=========================================================*/
/* end a timer. return the number of seconds. pass the count
 * from TimerStart()                                       */
double TimerEnd(my_timer s)
{
#ifdef X86_LINUX_TIMER
	return (double)(rdtsc() - s) * secs_per_clock;
#else
	my_timer e;
	
	gettimeofday(&e, NULL);
	return TimeDiff(&s, &e);
#endif
}

/*=========================================================*/
/* wait for a certain interval to pass, spinning. make sure
 * to reset the timer after returning from this            */
int TimerSleep(my_timer s, my_timer interval)
{
	my_timer trigger;
	int dropped;

#ifdef X86_LINUX_TIMER
	trigger = interval + s;

	if (trigger < rdtsc()) 
	{
		dropped = (rdtsc() - trigger) / (trigger - s) + 1;

		trigger += dropped*interval;
		while (rdtsc() < trigger);
		return dropped;
	}

	while (rdtsc() < trigger);

#else
	my_timer now;

	trigger.tv_usec = (s.tv_usec + interval.tv_usec) % (int)(1e6);
	trigger.tv_sec = s.tv_sec + interval.tv_sec + (s.tv_usec + interval.tv_usec) / (int)(1e6);

	gettimeofday(&now, NULL);
	if ((trigger.tv_sec < now.tv_sec) && (trigger.tv_usec < now.tv_usec))
	{
		dropped = TimeDiff(&now, &trigger) / TimeDiff(&trigger, &s) + 1;
		trigger.tv_usec += (dropped * interval.tv_usec) % (int)(1e6);
		trigger.tv_sec += dropped*interval.tv_sec + (dropped*interval.tv_usec) / 1e6;

		while (1)
		{
			gettimeofday(&now, NULL);
			if ((now.tv_sec >= trigger.tv_sec) && (now.tv_usec >= trigger.tv_usec))
				return dropped;
		}
	}

	while (1)
	{
		gettimeofday(&now, NULL);
		if ((now.tv_sec >= trigger.tv_sec) && (now.tv_usec >= trigger.tv_usec))
			return 0;
	}

#endif

	return 0;

}

/*=========================================================*/
/* wait for a certain interval to pass, spinning. make sure
 * to reset the timer after returning from this            
 * then reset the timer                                    */
int TimerSleepAndReset(my_timer *s, my_timer interval)
{
	my_timer trigger;
	int dropped;

#ifdef X86_LINUX_TIMER
	trigger = interval + *s;

	if (trigger < rdtsc()) 
	{
		dropped = (rdtsc() - trigger) / (trigger - *s) + 1;

		trigger += dropped*interval;
		while (rdtsc() < trigger);
		*s = rdtsc();
		return dropped;
	}

	while (rdtsc() < trigger);
	*s = rdtsc();
#else

	my_timer now;

	trigger.tv_usec = (s->tv_usec + interval.tv_usec) % (int)(1e6);
	trigger.tv_sec = s->tv_sec + interval.tv_sec + (s->tv_usec + interval.tv_usec) / (int)(1e6);

	gettimeofday(&now, NULL);
	if ((now.tv_sec > trigger.tv_sec) || 
				((now.tv_sec == trigger.tv_sec) && (now.tv_usec >= trigger.tv_usec)))
	{
		dropped = TimeDiff(&now, &trigger) / TimeDiff(&trigger, s) + 1;
		trigger.tv_usec += (dropped * interval.tv_usec) % (int)(1e6);
		trigger.tv_sec += dropped*interval.tv_sec + (dropped*interval.tv_usec) / (int)(1e6);

		while (1)
		{
			gettimeofday(&now, NULL);
			if ((now.tv_sec > trigger.tv_sec) || 
				((now.tv_sec == trigger.tv_sec) && (now.tv_usec >= trigger.tv_usec)))

			{
				gettimeofday(s, NULL);
				return dropped;
			}
		}
	}

	while (1)
	{
		gettimeofday(&now, NULL);

		if ((now.tv_sec > trigger.tv_sec) || 
				((now.tv_sec == trigger.tv_sec) && (now.tv_usec >= trigger.tv_usec)))
		{
			gettimeofday(s, NULL);
			return 0;
		}
	}

#endif

	return 0;
}


